#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
[Anonymous1]
Modified from plot-point.py used in tensorpack (https://github.com/ppwwyyxx/tensorpack).
Usage:
python3 scripts/my_plot_point.py -i $PKL_FILE ...
The format of the $PKL_FILE should be (xs, ys)
where xs is a list of lists, e.g. [[1,2,3]] or [[1,2], [2,4]]
and ys is a list of dicts, e.g. [{'avg':[1.0, 2.0]}, {'avg':[1.5, 0.5]}]
The elements in the dict are list of data point for each keys (e.g. avg/std)
----------------------------------------------------------------------------------------

A general curve plotter to create curves such as:
https://github.com/ppwwyyxx/tensorpack/tree/master/examples/ResNet

A simplest example:
$ cat examples/train_log/mnist-convnet/stat.json \
        | jq '.[] | .train_error, .validation_error' \
        | paste - - \
        | plot-point.py --legend 'train,val' --xlabel 'epoch' --ylabel 'error'

For more usage, see `plot-point.py -h` or the code.
"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.font_manager as fontm
import argparse
import copy
import pickle
import sys
from collections import defaultdict
from itertools import chain
import six

# from matplotlib import rc
# rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})
# rc('font',**{'family':'sans-serif','sans-serif':['Microsoft Yahei']})
# rc('text', usetex=True)

STDIN_FNAME = '-'


def get_args():
    description = "plot points into graph."
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('-i', '--input',
                        help='input data file, use "-" for stdin. Default stdin. Input \
            format is many rows of DELIMIETER-separated data',
                        default='-')
    parser.add_argument('-o', '--output',
                        help='output image', default='')
    parser.add_argument('--show',
                        help='show the figure after rendered',
                        action='store_true')
    parser.add_argument('-c', '--column',
                        help="describe each column in data, for example 'x,y,y'. \
            Default to 'y' for one column and 'x,y' for two columns. \
            Plot attributes can be appended after 'y', like 'ythick;cr'. \
            By default, assume all columns are y. \
            ")
    parser.add_argument('-t', '--title',
                        help='title of the graph',
                        default='')
    parser.add_argument('--xlabel',
                        help='x label', type=six.text_type)
    parser.add_argument('--ylabel',
                        help='y label', type=six.text_type)
    parser.add_argument('--xlim',
                        help='x lim', type=float, nargs=2)
    parser.add_argument('--ylim',
                        help='y lim', type=float, nargs=2)
    parser.add_argument('-s', '--scale',
                        help='scale of each y, separated by comma')
    parser.add_argument('--annotate-maximum',
                        help='annonate maximum value in graph',
                        action='store_true')
    parser.add_argument('--annotate-minimum',
                        help='annonate minimum value in graph',
                        action='store_true')
    parser.add_argument('--xkcd',
                        help='xkcd style',
                        action='store_true')
    parser.add_argument('--decay',
                        help='exponential decay rate to smooth Y',
                        type=float, default=0)
    parser.add_argument('-l', '--legend',
                        help='legend for each y')
    parser.add_argument('-d', '--delimeter',
                        help='column delimeter', default='\t')
    parser.add_argument('-mm', '--plot_minmax',  action='store_true',
                        help='plot minmax')

    global args
    args = parser.parse_args()

    if not args.show and not args.output:
        args.show = True


def filter_valid_range(points, rect):
    """rect = (min_x, max_x, min_y, max_y)"""
    ret = []
    for x, y in points:
        if x >= rect[0] and x <= rect[1] and y >= rect[2] and y <= rect[3]:
            ret.append((x, y))
    if len(ret) == 0:
        ret.append(points[0])
    return ret


def exponential_smooth(data, alpha):
    """ smooth data by alpha. returned a smoothed version"""
    ret = np.copy(data)
    now = data[0]
    for k in range(len(data)):
        ret[k] = now * alpha + data[k] * (1 - alpha)
        now = ret[k]
    return ret


def annotate_min_max(data_x, data_y, ax):
    max_x, min_x = max(data_x), min(data_x)
    max_y, min_y = max(data_y), min(data_y)
    x_range = max_x - min_x
    y_range = max_y - min_y
    x_max, y_max = data_y[0], data_y[0]
    x_min, y_min = data_x[0], data_y[0]

    for i in range(1, len(data_x)):
        if data_y[i] > y_max:
            y_max = data_y[i]
            x_max = data_x[i]
        if data_y[i] < y_min:
            y_min = data_y[i]
            x_min = data_x[i]

    rect = ax.axis()
    if args.annotate_maximum:
        text_x, text_y = filter_valid_range([
            (x_max + 0.05 * x_range,
                y_max + 0.025 * y_range),
            (x_max - 0.05 * x_range,
                y_max + 0.025 * y_range),
            (x_max + 0.05 * x_range,
                y_max - 0.025 * y_range),
            (x_max - 0.05 * x_range,
                y_max - 0.025 * y_range)],
            rect)[0]
        ax.annotate('maximum ({:d},{:.3f})' . format(int(x_max), y_max),
                    xy=(x_max, y_max),
                    xytext=(text_x, text_y),
                    arrowprops=dict(arrowstyle='->'))
    if args.annotate_minimum:
        text_x, text_y = filter_valid_range([
            (x_min + 0.05 * x_range,
                y_min - 0.025 * y_range),
            (x_min - 0.05 * x_range,
                y_min - 0.025 * y_range),
            (x_min + 0.05 * x_range,
                y_min + 0.025 * y_range),
            (x_min - 0.05 * x_range,
                y_min + 0.025 * y_range)],
            rect)[0]
        ax.annotate('minimum ({:d},{:.3f})' . format(int(x_min), y_min),
                    xy=(x_min, y_min),
                    xytext=(text_x, text_y),
                    arrowprops=dict(arrowstyle='->'))
        # ax.annotate('{:.3f}' . format(y_min),
        # xy = (x_min, y_min),
        # xytext = (text_x, text_y),
        # arrowprops = dict(arrowstyle = '->'))


def plot_args_from_column_desc(desc):
    # if not desc:
    #     return {}
    ret = {}
    desc = desc.split(';')
    if True: #'thick' in desc:
        ret['lw'] = 3
    if 'dash' in desc:
        ret['ls'] = '--'
    for v in desc:
        if v.startswith('c'):
            ret['color'] = v[1:]
    return ret


def do_plot(data_xs, data_ys):
    """
    data_xs: list of 1d array, either of size 1 or size len(data_ys)
    data_ys: list of dict, each key contains 1d array
    """
    fig = plt.figure(figsize=(16.18 / 1.2, 10 / 1.2))
    ax = fig.add_axes((0.1, 0.2, 0.8, 0.7))
    nr_y = len(data_ys)
    y_column = args.y_column

    # parse legend and y-scale
    if args.legend:
        legends = args.legend.split(',')
        assert len(legends) == nr_y
    else:
        legends = None  # range(nr_y) #None
    if args.scale:
        scale = map(float, args.scale.split(','))
        assert len(scale) == nr_y
    else:
        scale = [1.0] * nr_y

    minx, maxx = None, None
    for yidx in range(nr_y):
        plotargs = plot_args_from_column_desc(y_column[yidx][1:])
        now_scale = scale[yidx]
        data_y = data_ys[yidx]
        # print(data_y['avg'])
        assert type(data_y) is dict or type(data_y) is defaultdict
        assert 'avg' in data_y

        for key in data_y.keys():
            data_y[key] = np.array(data_y[key])
            data_y[key] = data_y[key] * now_scale
        leg = legends[yidx] if legends else None
        if now_scale != 1:
            leg = "{}*{}".format(now_scale if int(now_scale) != now_scale else int(now_scale), leg)
        data_x = data_xs[0] if len(data_xs) == 1 else data_xs[yidx]
        if minx is None:
            minx, maxx = min(data_x), max(data_x)
        else:
            minx, maxx = min(minx, min(data_x)), max(maxx, max(data_x))
        assert len(data_x) >= len(data_y['avg']), \
            "x column is shorter than y column! {} < {}".format(
            len(data_x), len(data_y['avg']))
        truncate_data_x = data_x[:len(data_y['avg'])]
        p = plt.plot(truncate_data_x, data_y['avg'], label=leg, **plotargs)

        c = p[0].get_color()
        if args.plot_minmax:
            if 'min' in data_y:
                min_plotargs = copy.copy(plotargs)
                min_plotargs['ls'] = ':'
                min_plotargs['color'] = c
                p_min = plt.plot(truncate_data_x, data_y['min'], **min_plotargs)

            if 'max' in data_y:
                max_plotargs = copy.copy(plotargs)
                # max_plotargs['lw'] = 5
                max_plotargs['ls'] = '--'
                max_plotargs['color'] = c
                p_min = plt.plot(truncate_data_x, data_y['max'], **max_plotargs)

        if 'std' in data_y:
            std = data_y['std']
            avg = data_y['avg']
            plt.fill_between(truncate_data_x, avg - std, avg + std, alpha=0.1, facecolor=c)

        # plt.fill_between(truncate_data_x, data_y['avg'], alpha=0.1, facecolor=c)

        # ax.set_aspect('equal', 'datalim')
        # ax.spines['right'].set_color('none')
        # ax.spines['left'].set_color('none')
        # plt.xticks([])
        # plt.yticks([])

        if args.annotate_maximum or args.annotate_minimum:
            annotate_min_max(truncate_data_x, data_y['avg'], ax)

    if args.xlabel:
        plt.xlabel(args.xlabel, fontsize='20')
    if args.ylabel:
        plt.ylabel(args.ylabel, fontsize='20')
    if args.xlim:
        plt.xlim(args.xlim[0], args.xlim[1])
    if args.ylim:
        plt.ylim(args.ylim[0], args.ylim[1])
    plt.legend(loc='best', fontsize='20')

    # adjust maxx
    # minx, maxx = min(data_x), max(data_x)
    new_maxx = maxx + (maxx - minx) * 0.05
    plt.xlim(minx, new_maxx)

    for label in chain.from_iterable(
            [ax.get_xticklabels(), ax.get_yticklabels()]):
        label.set_fontproperties(fontm.FontProperties(size=15))

    ax.grid(color='gray', linestyle='dashed')
    # ax.set_yscale('log')

    plt.title(args.title, fontdict={'fontsize': '30'})

    if args.output != '':
        plt.savefig(args.output, bbox_inches='tight')
    if args.show:
        plt.show()


def main():
    get_args()
    # parse input args
    with open(args.input, 'rb') as f:
        data_xs, data_ys = pickle.load(f)
    nr_column = len(data_ys)
    # print(data_ys)

    if args.column is None:
        column = ['y'] * nr_column
    else:
        column = args.column.strip().split(',')
    for k in column:
        assert k[0] in ['x', 'y']
    assert nr_column == len(column), "Column and data doesn't have same length. {}!={}".format(nr_column, len(column))
    args.y_column = [v for v in column if v[0] == 'y']
    args.y_column_idx = [idx for idx, v in enumerate(column) if v[0] == 'y']
    args.x_column = [v for v in column if v[0] == 'x']
    args.x_column_idx = [idx for idx, v in enumerate(column) if v[0] == 'x']
    nr_x_column = len(args.x_column)
    nr_y_column = len(args.y_column)
    if nr_x_column > 1:
        assert nr_x_column == nr_y_column, \
            "If multiple x columns are used, nr_x_column must equals to nr_y_column"

    for idx, data_y in enumerate(data_ys):
        if args.decay != 0:
            for k in data_y.keys():
                if len(data_y[k] > 0):
                    data_ys[idx][k] = exponential_smooth(data_y[k], args.decay)
    do_plot(data_xs, data_ys)


if __name__ == '__main__':
    main()


    # data_xs = [[0,1,2,3]]
    # def get_rand_y():
    #     y = dict(avg=[], min=[], max=[], std=[])
    #     for i in range(4):
    #         a = np.random.randint(10, size=10)
    #         y['avg'].append(np.mean(a))
    #         y['min'].append(np.min(a))
    #         y['max'].append(np.max(a))
    #         y['std'].append(np.std(a))
    #     for k in y.keys():
    #         y[k] = np.array(y[k])
    #     return y

    # y1, y2 = get_rand_y(), get_rand_y()

    # data_ys = [y1, y2]
    # dict(min=np.array([1,2,3,4]), avg=np.array([5, 10, 15, 20]), max=np.array([6,11,16,21]), std=np.array([0.1,1,2,3]))]
